/*
 * rpicd.c
 *
 *  Created on: 10/04/2013
 *      Author: Manuel Serrano
 *      http://tren.enmicasa.net
 */

/*
    rpic daemin to interface rpic in Raspberry PI
*/

#define RPICDPORT 8788
#define MAXNCLIENTS 3
#define RPISERIAL "/dev/ttyAMA0"
#define RPIBAUD 115200
#define LEDREADY  3
#define BUTTONPIN 0
#define SEC4HALT  3
#define LOGDIR	"/var/log/"
#define LOGFILE	"rpicd.log"
#define PIDDIR  "/opt/rocrail/"
#define PIDFILE "rpicd.pid"

#include <stdio.h>
//#include <string.h>    //strlen
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <arpa/inet.h> //inet_addr
#include <pthread.h>   //for threading , link with lpthread
#include <time.h>
#ifdef __arm__
#include <wiringPi.h>
#include <wiringSerial.h>
#endif


void resetbutton (void);
int createSocket(void);
int initWiringPi(void);
//the thread function
void *connection_handler(void *);


static volatile int  serial_desc;			//Only one serial port can be active...only one can write into the uart
static volatile struct timeval start, stop;
static FILE *flog= NULL;					//only one log file

int main(int argc , char *argv[])
{
	//Following http://www.netzmafia.de/skripten/unix/linux-daemon-howto.html

	/* Our process ID and Session ID */
	pid_t pid, sid;

    /* Fork off the parent process */
    pid = fork();
    if (pid < 0) {
        exit(EXIT_FAILURE);
    	//return -1;
    }

    /* If we got a good PID, then
       we can exit the parent process. */
    if (pid > 0) {
        exit(EXIT_SUCCESS);
    	//return 0;
    }

    /* Change the file mode mask */
    umask(0);


    /* Create a new SID for the child process */
	sid = setsid();
	if (sid < 0) {
		exit(EXIT_FAILURE);
		//return -1;
	}

	/* Change the current working directory and creadte a pid file*/
	if ((chdir(PIDDIR)) < 0) {
		exit(EXIT_FAILURE);
		//return -1;
	}

	FILE *fpid= NULL;
	if ((fpid = fopen (PIDFILE, "w+"))<0){
    	return -1;
    }
	fprintf(fpid, "%i",(int)sid);
	fclose (fpid);


	/* Change the current working directory for lof*/
	if ((chdir(LOGDIR)) < 0) {
		//exit(EXIT_FAILURE);
		return -1;
	}

   /* Open any logs here */
    if ((flog = fopen (LOGFILE, "w+"))<0){
    	return -1;
    }

    fprintf(flog, "*****************Logging info rpicd********************\n");
    fflush(flog);

	/* Close out the standard file descriptors */
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);

	/* Daemon-specific initialization goes here */

	int socket_desc, client_sock , c , *new_sock;
    struct sockaddr_in client;

    if (initWiringPi()<0){
    	fclose(flog);
    	return -1;
    }

    if ((socket_desc=createSocket())<0){
    	fclose(flog);
    	return -1;
    }
    //Listen
    listen(socket_desc , MAXNCLIENTS);

    //Accept and incoming connection
    //puts("Waiting for incoming connections...");
    fprintf(flog, "Waiting for incoming connections...\n");
    fflush(flog);
    c = sizeof(struct sockaddr_in);

    /* The Big Loop */

    while( (client_sock = accept(socket_desc, (struct sockaddr *)&client, (socklen_t*)&c)) )
    {
        //puts("Connection accepted");
    	fprintf(flog, "Connection accepted.\n");
    	fflush(flog);
#ifdef __arm__
        digitalWrite (LEDREADY, 1) ;
#endif
        pthread_t sniffer_thread;
        new_sock = malloc(1);
        *new_sock = client_sock;

        if( pthread_create( &sniffer_thread , NULL ,  connection_handler , (void*) new_sock) < 0)
        {
        	fprintf(flog, "ERROR: Could not create thread.\n");
        	fclose(flog);
        	//perror("could not create thread");
            return -1;
        }

        //Now join the thread , so that we dont terminate before the thread
        //pthread_join( sniffer_thread , NULL);
        //puts("New Handler assigned");
        fprintf(flog, "New Handler assigned.\n");
        fflush(flog);
    }

    if (client_sock < 0)
    {
        //perror("accept failed");
    	fprintf(flog, "ERROR: Accept failed.\n");
    	fclose(flog);
        return -1;
     }
    fclose (flog);
    return 0;
}

int createSocket(){
	int socket_desc;

    struct sockaddr_in server;

    //Create socket
    socket_desc = socket(AF_INET , SOCK_STREAM , 0);
    if (socket_desc == -1)
    {
        //perror("Could not create socket");
    	fprintf(flog, "ERROR: Could not create socket.\n");
        return -1;
    }
    fprintf(flog, "Socket created.\n");
    fflush(flog);
    //puts("Socket created");

    //Prepare the sockaddr_in structure
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = htons( RPICDPORT );

    //Bind
    if( bind(socket_desc,(struct sockaddr *)&server , sizeof(server)) < 0)
    {
        //print the error message
        //perror("bind failed. Error");
    	fprintf(flog, "ERROR: Bind failed.\n");
        return -1;
    }
    fprintf(flog, "Binds done.\n");
    fflush(flog);
    return socket_desc;

}

int initWiringPi(){
#ifdef __arm__
	  if ((serial_desc = serialOpen (RPISERIAL, RPIBAUD)) < 0)
	  {
		  //perror("Unable to open serial port") ;
	      fprintf(flog, "ERROR: Unable to open serial port.\n");
	    return -1 ;
	  }

	  if (wiringPiSetup () == -1)
	  {
	    //perror ("Unable to start wiringPi") ;
		fprintf(flog, "ERROR: Unable to start wiringPi.\n");
	    return -1 ;
	  }

	 if (wiringPiISR (BUTTONPIN, INT_EDGE_BOTH, &resetbutton) < 0)
	 {
		 //perror ("Unable to set up the buttons");
		 fprintf(flog, "ERROR: Unable to set up the buttons.\n");
		return -1;
	 } else {
		 //puts ("Reset button ready");
		 fprintf(flog, "Reset button ready.\n");
		 fflush(flog);
	}

	  pinMode (LEDREADY, OUTPUT) ;

#endif
	return 0;
}

void resetbutton (void)
{
#ifdef __arm__
  int readPin;
  int sec;
  readPin=digitalRead (BUTTONPIN);

  if (!readPin){
    gettimeofday(&start, NULL);
  }else{
    gettimeofday(&stop, NULL);
    sec=(int)(stop.tv_sec-start.tv_sec);
    if (sec>SEC4HALT) {
            //puts ("Halt requested");
    	     fprintf(flog, "Halt requested.\n");
    	     fclose(flog);
            system("sudo halt");
          } else {
        	//puts ("Reset requested");
        	fprintf(flog, "Reset requested.\n");
        	fflush(flog);
            system("sudo /etc/init.d/rocraild reset");
            //should it be necessary to send a message to everyone.. or check it is just a localhost?
          }
  }

#endif

}

/*
 * This will handle connection for each client
 * */
void *connection_handler(void *socket_desc)
{
    //Get the socket descriptor
    int sock = *(int*)socket_desc;
    int read_size;
    unsigned char i;
    char client_message[10];

    //Receive a message from client
    while( (read_size = recv(sock , client_message , 10 , 0)) > 0 )
    {
    	for (i = 0; i < read_size; i++) {
    	      //printf("Command sent buffer %d: %c %X \n", i, client_message[i],client_message[i]);
    		fprintf(flog, "Command sent buffer %d: %X \n", i, client_message[i]);
#ifdef __arm__
    	      serialPutchar (serial_desc,client_message[i]);
#endif
    	}
    	fflush(flog);
    }

    if(read_size == 0)
    {
    	//puts("Client disconnected");
        //fflush(stdout);
    	fprintf(flog, "Client disconnected\n");
    	fflush(flog);
 #ifdef __arm__
        digitalWrite (LEDREADY, 0) ;		//Let's see
#endif
    }
    else if(read_size == -1)
    {
        //perror("recv failed");
    	fprintf(flog, "ERROR: recv failed\n");
    	fflush(flog);
#ifdef __arm__
        digitalWrite (LEDREADY, 0) ;		//Let's see
#endif
    }
    //Free the socket pointer
    free(socket_desc);
    return 0;
}



